<?php

class postgresql_fts_views_handler_filter extends views_handler_filter {
  /**
   * Inform the system about the various option that we
   * set for this PostgreSQL filter handler, appending to
   * the options set by the parent class.
   */
  function option_definition() {
    $options = parent::option_definition();

    $options['expose']['contains']['required'] = array(
      'default' => FALSE,
      'bool' => TRUE,
    );

    // Add the column option so it can be saved afterwards.
    $options['column'] = array('default','default');

    return $options;
  }

  /**
   * Build the specific options needed for this PostgreSQL
   * FTS filter, appending to the options set by the parent class.
   */
  function options_form(&$form, &$form_state) {
    parent::options_form($form, $form_state);

    $form['value'] = array(
        '#type' => 'textfield',
        '#title' => t('Search string'),
        '#description' => t('Enter your Full Text search string.'),
        '#size' => 60,
        '#default_value' => $this->value,
      );

    $form['column'] = array(
      '#type' => 'select',
      '#title' => t('Indexed columns'),
      '#description' => t('Select on which indexed columns you want to perform an FTS search.'),
      '#options' => array_flip($this->definition['data']['columns']),
      '#default_value' => $this->options['column'],
    );

    // Remove some default options from the form
    // that don't make sense for this handler.
    unset($form['expose']['multiple']);
    unset($form['expose']['remember']);
    unset($form['expose']['remember_roles']);
  }

  function value_form(&$form, &$form_state) {
    parent::value_form($form, $form_state);

    $form['value'] = array(
        '#type' => 'textfield',
        '#title' => t('Search string'),
        '#description' => t('Enter your Full Text search string.'),
        '#size' => 60,
        '#default_value' => $this->value,
      );
  }

  /**
   * Set the different operators that PostgreSQL
   * FTS system supports.
   */
  function operators() {
    $operators = array(
      'plain' => array(
        'title' => t('Default full text'),
        'short' => t('Plain'),
        'method' => 'op_plain',
        'values' => 1,
      ),
    );

    return $operators;
  }

  /**
   * Build the strings that represent the different operator
   * options in the operators radio button list rendered
   * by the parent class.
   */
  function operator_options($which = 'title') {
    $options = array();
    foreach ($this->operators() as $id => $info) {
      $options[$id] = $info[$which];
    }

    return $options;
  }


  function operator_values($values = 1) {
    $options = array();
    foreach ($this->operators() as $id => $info) {
      if (isset($info['values']) && $info['values'] == $values) {
        $options[] = $id;
      }
    }

    return $options;
  }

  /**
   * Prepare a string to display in the overview screen,
   * next to the filter.
   */
  function admin_summary() {
    if ($this->is_a_group()) {
      return t('grouped');
    }
    if (!empty($this->options['exposed'])) {
      return t('exposed');
    }

    $options = $this->operator_options('short');
    $output = '';
    if (!empty($options[$this->operator])) {
      $output = check_plain($options[$this->operator]);
    }
    if (in_array($this->operator, $this->operator_values(1))) {
      $output .= ' ' . check_plain($this->value);
    }
    return $output;
  }

  /**
   * Add this filter to our query.
   */
  function query() {
    $this->ensure_my_table();

    $info = $this->operators();
    if (!empty($info[$this->operator]['method'])) {
      $this->{$info[$this->operator]['method']}();
    }
  }

  /**
   * Perform the actual PostgreSQL FTS search based on the default
   * '@@' operator, joining the desired FTS index table to make
   * its columns available during the WHERE filtering.
   */
  function op_plain() {
    // Fetch the index data patched through by the filter handler creator.
    $data = $this->definition['data'];

    // Assign the needed query settings.
    $placeholder = $this->placeholder();
    $entity_base_table = $data['entity_base_table'];
    $entity_base_table_alias = $this->ensure_my_table($entity_base_table);
    $entity_id_field = $data['entity_id_field'];
    $fts_table = $data['fts_table'];
    $fts_column = 'entity_id';
    $fts_table_alias = 'pgfts';

    // Prepare to join the FTS index table to the entity table.
    $join = new views_join;
    $join->left_table = $entity_base_table_alias;
    $join->left_field = $entity_id_field;
    $join->table = $fts_table;
    $join->field = $fts_column;
    $join->type = 'LEFT';

    // Do the join.
    $join_alias = $this->query->add_relationship($fts_table_alias, $join, $entity_base_table);

    // Perform the actual Full Text filtering by
    // adding a where clause base on the join we did above.
    $column = $this->options['column'];
    $this->query->add_where_expression($this->options['group'], "$join_alias.$column @@ plainto_tsquery($placeholder)", array($placeholder => $this->value));
  }
}